/*
 * Flipdot Demo serial control
 * Copyright (C) 2013  B. Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <QtGui>
#include <QSocketNotifier>
#include <QMessageBox>
#include <QSlider>
#include <QComboBox>
#include <QPushButton>
#include <QCheckBox>
#include <QColor>
#include <QUdpSocket>

#include <stdint.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#ifndef __WIN32__
#include <termios.h>
#endif

#include "dialog.h"
#include "helpdia.h"
#include "fdbox.h"
#include "logo.h"
#include "vagrearg.logo-563x160-transparent.xpm"

#define REENUMERATETIMER	2000
#define KEEPALIVETIMER		10000
#define RUNTIMER		25

static void mylogger(const char *fmt, ...)
{
#ifdef __WIN32__
	(void)fmt;
#else
	va_list va;
	va_start(va, fmt);
	vfprintf(stderr, fmt, va);
	va_end(va);
#endif
}

Dialog::Dialog()
{
	portisopen = false;	/* Set when serial port is attached */

	/* Re-enumerate the list automagically */
	reenumtmr = new QTimer(this);
	connect(reenumtmr, SIGNAL(timeout()), this, SLOT(enumports()));
	reenumtmr->start(REENUMERATETIMER);

	/* Keepalive timer when auto-flip-program is disabled */
	autotmr = new QTimer(this);
	connect(autotmr, SIGNAL(timeout()), this, SLOT(fliptimeout()));
	autotmr->stop();

	runtmr = new QTimer(this);
	connect(runtmr, SIGNAL(timeout()), this, SLOT(runflip()));
	runtmr->stop();

	setMinimumWidth(400);

	/* The close button */
	QDialogButtonBox *buttonBox = new QDialogButtonBox(QDialogButtonBox::Close|QDialogButtonBox::Help);
	connect(buttonBox, SIGNAL(rejected()), this, SLOT(reject()));
	connect(buttonBox, SIGNAL(helpRequested()), this, SLOT(dohelp()));

	connect(this, SIGNAL(finished(int)), this, SLOT(doclose(int)));

	/* Load file button */
	loader = new QPushButton("Load Flip-file");
	buttonBox->addButton(loader, QDialogButtonBox::ActionRole);
	connect(loader, SIGNAL(clicked(bool)), this, SLOT(loadfile(bool)));

	/* Run flip data button */
	runner = new QPushButton("Start Flip-file");
	buttonBox->addButton(runner, QDialogButtonBox::ActionRole);
	connect(runner, SIGNAL(clicked(bool)), this, SLOT(runfile(bool)));
	runner->setEnabled(false);

	/* Serial port selection */
	QHBoxLayout *hbox = new QHBoxLayout();
	hbox->addWidget(new QLabel("Serial Port"));
	QPushButton *b;
	hbox->addWidget(b = new QPushButton("Refresh list"));
	connect(b, SIGNAL(clicked(bool)), this, SLOT(enumports(bool)));
	port = new QComboBox();
	port->setEditable(false);
	connect(port, SIGNAL(currentIndexChanged(int)), this, SLOT(setport(int)));
	hbox->addWidget(port);

	/* Auto-flip-program on/off control */
	QHBoxLayout *cbox = new QHBoxLayout();
	cbox->addWidget(new QLabel("Stop automatic flip-program when attached:"));
	autoflip = new QCheckBox();
	autoflip->setCheckState(Qt::Checked);
	cbox->addWidget(autoflip);
	connect(autoflip, SIGNAL(stateChanged(int)), this, SLOT(checker(int)));

	/* The flipdots */
	QHBoxLayout *hbf = new QHBoxLayout();
	for(int i = 0; i < 7; i++) {
		flipdots[i] = new fdbox('A'+i, 'a'+i);
		connect(flipdots[i], SIGNAL(changed(int)), this, SLOT(doflip(int)));
		hbf->addWidget(flipdots[i]);
	}

	QLabel *lab = new QLabel("<a href=\"http://www.flipdots.com\">http://www.flipdots.com</a>");
	lab->setOpenExternalLinks(true);
	lab->setTextFormat(Qt::RichText);
	lab->setScaledContents(true);
	logo *lgp = new logo();
	QPixmap pm(vagrearg_logo_563x160_transparent_xpm);
	lgp->setPixmap(pm.scaled(141, 40, Qt::IgnoreAspectRatio, Qt::SmoothTransformation));

	QHBoxLayout *hdr = new QHBoxLayout();
	hdr->addWidget(lgp, 0, Qt::AlignLeft);
	hdr->addWidget(lab, 0, Qt::AlignRight);

	/* All in a layout */
	QVBoxLayout *mainLayout = new QVBoxLayout();
	mainLayout->addLayout(hdr);
	QFrame *f = new QFrame();
	f->setFrameStyle(QFrame::HLine | QFrame::Raised);
	f->setLineWidth(3);
	mainLayout->addWidget(f);
	mainLayout->addLayout(hbox);
	mainLayout->addLayout(cbox);
	f = new QFrame();
	f->setFrameStyle(QFrame::HLine | QFrame::Raised);
	f->setLineWidth(3);
	mainLayout->addWidget(f);
	mainLayout->addLayout(hbf, 1);
	f = new QFrame();
	f->setFrameStyle(QFrame::HLine | QFrame::Raised);
	f->setLineWidth(3);
	mainLayout->addWidget(f);
	mainLayout->addWidget(buttonBox);
	setLayout(mainLayout);

	setWindowTitle("FlipDemo - Flipdot Control");

	enumports(true);	/* Start by enumerating all serial ports */
	filedir = QString(".");
}

void Dialog::dohelp()
{
	HelpDia hd;
	hd.exec();
}

void Dialog::doclose(int)
{
	if(portisopen) {
		if(autoflip->checkState() == Qt::Checked) {
			ssize_t err = dowrite("+ ", 2);
			if(err != 2)
				mylogger("Failed to disable flipprogram\n");
		}
		closeport();
	}
}

void Dialog::runfile(bool)
{
	if(!flipdata.size())
		return;
	if(runtmr->isActive()) {
		runtmr->stop();
		runner->setText("Start Flip-file");
	} else {
		flipdatapos = 0;
		runtmr->start(RUNTIMER);
		runner->setText("Stop Flip-file");
	}
}

void Dialog::runflip()
{
	if(flipdatapos >= flipdata.size())
		flipdatapos = 0;
	while(flipdatapos < flipdata.size()) {
		char c = flipdata[flipdatapos];
		flipdatapos++;
		if(strchr("abcdefgABCDEFG", c)) {
			doflip(c);
			if(islower(c))
				flipdots[c-'a']->setstate(0);
			else
				flipdots[c-'A']->setstate(1);
			break;
		}
		if(c == '.' || c == ',')
			break;
		if(c == 'q' || c == 'Q')
			runfile(false);
	}
}

void Dialog::loadfile(bool)
{
	QString fn = QFileDialog::getOpenFileName(this, "Load flip-file", filedir, "Any File (*.* *);;Text Files (*.txt)");
	if(fn.isEmpty())
		return;
	QFile file(fn);
	if(!file.open(QIODevice::ReadOnly))
		return;
	if(file.size() > 512*1024*1024) {
		QMessageBox::warning(this, "File too big", QString("File ")+fn+" is too big (>512MByte)\nIt kinda defeats the purpose of limitation and constraint.");
		return;
	}
	flipdata = file.readAll();
	flipdatapos = 0;
	runner->setEnabled(flipdata.size() != 0);
	QDir dir;
	filedir = dir.cleanPath(fn + "/..");
}

void Dialog::checker(int st)
{
	if(!portisopen)
		return;
	if(st == Qt::Checked) {
		ssize_t err = dowrite("- ", 2);
		if(err != 2)
			mylogger("Failed to disable flipprogram\n");
		autotmr->start(KEEPALIVETIMER);
	} else if(st == Qt::Unchecked) {
		ssize_t err = dowrite("+ ", 2);
		if(err != 2)
			mylogger("Failed to enable flipprogram\n");
		autotmr->stop();
	}
}

void Dialog::fliptimeout()
{
	if(!portisopen)
		return;
	ssize_t err = dowrite("- ", 2);
	if(err != 2)
		mylogger("Autoflip disable keepalive failed\n");
}

void Dialog::doflip(int ch)
{
	char c = ch;
	if(!portisopen)
		return;
	ssize_t err = dowrite(&c, 1);
	if(err != 1)
		mylogger("write error: n=%d (errno=%d)\n", (int)err, errno);
	autotmr->stop();
	autotmr->start(KEEPALIVETIMER);
}

ssize_t Dialog::dowrite(const void *p, size_t n)
{
	if(!portisopen)
		return 0;
#ifdef __WIN32__
	DWORD res;
	BOOL rv = WriteFile(serfd, p, n, &res, NULL);
	return (ssize_t)(rv ? res : -1);
#else
	return ::write(serfd, p, n);
#endif
}

void Dialog::openport(const char *path)
{
	if(portisopen)
		closeport();
#ifdef __WIN32__
	DCB dcb;
	COMMTIMEOUTS cto = { 0, 0, 0, 0, 0 };
	serfd = CreateFileA(path, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, 0, 0);

	if(serfd == INVALID_HANDLE_VALUE) {
		QMessageBox::warning(this, "Serial port open", QString("Serial port ")+path+" cannot be opened. Maybe it was unplugged?\nYou may need to refresh the serial port list and try again.");
		return;
	}

	memset(&dcb, 0, sizeof(dcb));
	dcb.DCBlength = sizeof(dcb);
	dcb.BaudRate = CBR_1200;
	dcb.fBinary = 1;
	dcb.StopBits = ONESTOPBIT;
	dcb.ByteSize = 8;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	if(!SetCommMask(serfd, 0)) {
		QMessageBox::warning(this, "Serial port open", QString("Serial port ")+path+QString(" cannot configure settings (errno=%1)"".\nYou may need to refresh the serial port list and try again.").arg(GetLastError()));
		CloseHandle(serfd);
		return;
	}
	if(!SetCommTimeouts(serfd, &cto)) {
		QMessageBox::warning(this, "Serial port open", QString("Serial port ")+path+QString(" cannot configure timeouts (errno=%1)"".\nYou may need to refresh the serial port list and try again.").arg(GetLastError()));
		CloseHandle(serfd);
		return;
	}

	if(!SetCommState(serfd, &dcb)) {
		QMessageBox::warning(this, "Serial port open", QString("Serial port ")+path+QString(" cannot configure state (errno=%1)"".\nYou may need to refresh the serial port list and try again.").arg(GetLastError()));
		CloseHandle(serfd);
		return;
	}
#else
	struct termios tio;

	if(-1 == (serfd = ::open(path, O_RDWR|O_NOCTTY))) {
		/* Oops, device no longer available? */
		QMessageBox::warning(this, "Serial port open", QString("Serial port ")+path+" cannot be opened. Maybe it was unplugged?\nYou may need to refresh the serial port list and try again.");
		return;
	}
	if(-1 == tcgetattr(serfd, &tio)) {
		QMessageBox::warning(this, "Serial port open", QString("Serial port ")+path+QString(" cannot be read (errno=%1)"".\nYou may need to refresh the serial port list and try again.").arg(errno));
		::close(serfd);
		return;
	}
	cfmakeraw(&tio);
	cfsetspeed(&tio, B1200);
	tio.c_cflag &= ~CRTSCTS;
	tio.c_cflag |= CLOCAL;
	if(-1 == tcsetattr(serfd, TCSANOW, &tio)) {
		QMessageBox::warning(this, "Serial port open", QString("Serial port ")+path+QString(" cannot be configured (errno=%1)"".\nYou may need to refresh the serial port list and try again.").arg(errno));
		::close(serfd);
		return;
	}
#endif
	portisopen = true;
}

void Dialog::closeport()
{
	if(!portisopen)
		return;
#ifdef __WIN32__
	CloseHandle(serfd);
#else
	::close(serfd);
#endif
	portisopen = false;
}

void Dialog::setport(int v)
{
	if(portisopen) {
		for(int i = 0; i < 7; i++)
			flipdots[i]->setstate(0);
		ssize_t err = dowrite("a   b   c   d   e   f   g   ", 28);
		if(err != 28)
			mylogger("Failed to de-initialize all flipdots\n");
		if(autoflip->checkState() == Qt::Checked) {
			ssize_t err = dowrite("+ ", 2);
			if(err != 2)
				mylogger("Failed to enable flipprogram\n");
		}
		closeport();
	}
	reenumtmr->start(REENUMERATETIMER);
	autotmr->stop();

	if(!port->itemData(v).isNull()) {
		openport(port->itemData(v).toString().toLocal8Bit().constData());
		if(!portisopen)
			return;
		char c[8*7];
		memset(c, ' ', sizeof(c));
		if(autoflip->checkState() == Qt::Checked)
			c[0] = '-';
		for(int i = 0; i < 7; i++) {
			c[7*(i+1)] = flipdots[i]->getstate() ? 'A'+i : 'a'+i;
		}
		ssize_t err = dowrite(c, sizeof(c));
		if(err != sizeof(c))
			mylogger("Failed to initialize flipdots\n");
		reenumtmr->stop();
		if(autoflip->checkState() == Qt::Checked)
			autotmr->start(KEEPALIVETIMER);
	}
}

void Dialog::enumports(bool)
{
	enumports();
}

void Dialog::enumports()
{
	port->clear();
	port->addItem("-off-", QString());

#ifdef __WIN32__
	HKEY hkey;
	DWORD i = 0;

	if(ERROR_SUCCESS != RegOpenKeyExA(HKEY_LOCAL_MACHINE, "HARDWARE\\DEVICEMAP\\SERIALCOMM", 0, KEY_READ, &hkey))
		return;

	while(1) {
		char keyname[256];
		DWORD nkeyname = sizeof(keyname);
		char value[256];
		DWORD nvalue = sizeof(value);
		DWORD type;
		if(ERROR_SUCCESS != RegEnumValueA(hkey, i, keyname, &nkeyname, 0, &type, (LPBYTE)value, &nvalue))
			break;
		i++;
		if(type != REG_SZ)
			continue;
		port->addItem(QString(value), QString("\\\\.\\\\") + value);
	}
	RegCloseKey(hkey);
#else
	QString path("/dev/serial/by-id/");
	QDir d(path);
	if(d.exists()) {
		QStringList sl = d.entryList(QDir::NoDotAndDotDot | QDir::Files);
		for(int i = 0; i < sl.size(); i++) {
			char devpath[PATH_MAX];
			char *cptr;
			ssize_t len;
			if(-1 == (len = ::readlink((path + sl.at(i)).toLocal8Bit(), devpath, sizeof(devpath))))
				continue;
			devpath[sizeof(devpath)-1] = 0;
			if(len < (ssize_t)sizeof(devpath))
				devpath[len] = 0;
			if(!(cptr = strrchr(devpath, '/')))
				cptr = devpath;
			else
				cptr++;
			port->addItem(cptr, path + devpath);
		}
	}
#endif
}
